@@ -1,10 +1,9 @@
-# -*- coding: utf-8 -*-
 #
 # pylast -
 #     A Python interface to Last.fm and Libre.fm
 #
 # Copyright 2008-2010 Amr Hassan
-# Copyright 2013-2018 hugovk
+# Copyright 2013-2020 hugovk
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -20,65 +19,39 @@
 #
 # https://github.com/pylast/pylast
 
-from xml.dom import minidom, Node
 import collections
 import hashlib
+import html.entities
 import logging
 import shelve
 import ssl
-import sys
 import tempfile
 import time
+import warnings
 import xml.dom
-import html.entities
 from http.client import HTTPSConnection
+from urllib.parse import quote_plus
+from xml.dom import Node, minidom
 
-from . import version
+import pkg_resources
 
 __author__ = "Amr Hassan, hugovk, Mice Pápai"
-__copyright__ = (
-    "Copyright (C) 2008-2010 Amr Hassan, 2013-2018 hugovk, " "2017 Mice Pápai"
-)
+__copyright__ = "Copyright (C) 2008-2010 Amr Hassan, 2013-2020 hugovk, 2017 Mice Pápai"
 __license__ = "apache2"
 __email__ = "amr.hassan@gmail.com"
-__version__ = version.__version__
-
-if sys.version_info < (3,):
-    raise ImportError(
-        """pylast 3.0 and above are no longer compatible with Python 2.
-
-This is pylast {} and you are using Python {}.
-Make sure you have pip >= 9.0 and setuptools >= 24.2 and retry:
-
- $ pip install --upgrade pip setuptools
-
-Other choices:
-
-- Upgrade to Python 3.
-
-- Install an older version of pylast:
-
-$ pip install 'pylast<3.0'
-
-For more information:
-
-https://github.com/pylast/pylast/issues/265
-""".format(
-            version, ".".join([str(v) for v in sys.version_info[:3]])
-        )
-    )
-else:
-    # Keep importable on Python 2 for a while to show ImportError
-    from urllib.parse import quote_plus as url_quote_plus
+__version__ = pkg_resources.get_distribution(__name__).version
 
 
+# 1 : This error does not exist
 STATUS_INVALID_SERVICE = 2
 STATUS_INVALID_METHOD = 3
 STATUS_AUTH_FAILED = 4
 STATUS_INVALID_FORMAT = 5
 STATUS_INVALID_PARAMS = 6
 STATUS_INVALID_RESOURCE = 7
-STATUS_TOKEN_ERROR = 8
+# DeprecationWarning: STATUS_TOKEN_ERROR is deprecated and will be
+# removed in a future version. Use STATUS_OPERATION_FAILED instead.
+STATUS_OPERATION_FAILED = STATUS_TOKEN_ERROR = 8
 STATUS_INVALID_SK = 9
 STATUS_INVALID_API_KEY = 10
 STATUS_OFFLINE = 11
@@ -86,6 +59,20 @@ STATUS_SUBSCRIBERS_ONLY = 12
 STATUS_INVALID_SIGNATURE = 13
 STATUS_TOKEN_UNAUTHORIZED = 14
 STATUS_TOKEN_EXPIRED = 15
+STATUS_TEMPORARILY_UNAVAILABLE = 16
+STATUS_LOGIN_REQUIRED = 17
+STATUS_TRIAL_EXPIRED = 18
+# 19 : This error does not exist
+STATUS_NOT_ENOUGH_CONTENT = 20
+STATUS_NOT_ENOUGH_MEMBERS = 21
+STATUS_NOT_ENOUGH_FANS = 22
+STATUS_NOT_ENOUGH_NEIGHBOURS = 23
+STATUS_NO_PEAK_RADIO = 24
+STATUS_RADIO_NOT_FOUND = 25
+STATUS_API_KEY_SUSPENDED = 26
+STATUS_DEPRECATED = 27
+# 28 : This error is not documented
+STATUS_RATE_LIMIT_EXCEEDED = 29
 
 PERIOD_OVERALL = "overall"
 PERIOD_7DAYS = "7day"
@@ -201,10 +188,12 @@ class _Network:
         self.last_call_time = 0
         self.limit_rate = False
 
-        # Load session_key from authentication token if provided
+        # Load session_key and username from authentication token if provided
         if token and not self.session_key:
             sk_gen = SessionKeyGenerator(self)
-            self.session_key = sk_gen.get_web_auth_session_key(url=None, token=token)
+            self.session_key, self.username = sk_gen.get_web_auth_session_key_username(
+                url=None, token=token
+            )
 
         # Generate a session_key if necessary
         if (
@@ -876,7 +865,7 @@ class _Request:
         keys = list(self.params.keys())
         keys.sort()
 
-        cache_key = str()
+        cache_key = ""
 
         for key in keys:
             if key != "api_sig" and key != "api_key" and key != "sk":
@@ -906,13 +895,13 @@ class _Request:
 
         data = []
         for name in self.params.keys():
-            data.append("=".join((name, url_quote_plus(_string(self.params[name])))))
+            data.append("=".join((name, quote_plus(_string(self.params[name])))))
         data = "&".join(data)
 
         headers = {
             "Content-type": "application/x-www-form-urlencoded",
             "Accept-Charset": "utf-8",
-            "User-Agent": "pylast" + "/" + __version__,
+            "User-Agent": "pylast/" + __version__,
         }
 
         (host_name, host_subdir) = self.network.ws_server
@@ -1043,9 +1032,9 @@ class SessionKeyGenerator:
 
         return url
 
-    def get_web_auth_session_key(self, url, token=""):
+    def get_web_auth_session_key_username(self, url, token=""):
         """
-        Retrieves the session key of a web authorization process by its URL.
+        Retrieves the session key/username of a web authorization process by its URL.
         """
 
         if url in self.web_auth_tokens.keys():
@@ -1062,7 +1051,16 @@ class SessionKeyGenerator:
 
         doc = request.execute()
 
-        return doc.getElementsByTagName("key")[0].firstChild.data
+        session_key = doc.getElementsByTagName("key")[0].firstChild.data
+        username = doc.getElementsByTagName("name")[0].firstChild.data
+        return session_key, username
+
+    def get_web_auth_session_key(self, url, token=""):
+        """
+        Retrieves the session key of a web authorization process by its URL.
+        """
+        session_key, _username = self.get_web_auth_session_key_username(url, token)
+        return session_key
 
     def get_session_key(self, username, password_hash):
         """
@@ -1222,7 +1220,7 @@ class _Chartable:
         """
         Returns the weekly artist charts for the week starting from the
         from_date value to the to_date value.
-        Only for Tag or User.
+        Only for User.
         """
         return self.get_weekly_charts("artist", from_date, to_date)
 
@@ -1403,13 +1401,25 @@ class WSError(Exception):
             STATUS_INVALID_FORMAT = 5
             STATUS_INVALID_PARAMS = 6
             STATUS_INVALID_RESOURCE = 7
-            STATUS_TOKEN_ERROR = 8
+            STATUS_OPERATION_FAILED = 8
             STATUS_INVALID_SK = 9
             STATUS_INVALID_API_KEY = 10
             STATUS_OFFLINE = 11
             STATUS_SUBSCRIBERS_ONLY = 12
             STATUS_TOKEN_UNAUTHORIZED = 14
             STATUS_TOKEN_EXPIRED = 15
+            STATUS_TEMPORARILY_UNAVAILABLE = 16
+            STATUS_LOGIN_REQUIRED = 17
+            STATUS_TRIAL_EXPIRED = 18
+            STATUS_NOT_ENOUGH_CONTENT = 20
+            STATUS_NOT_ENOUGH_MEMBERS  = 21
+            STATUS_NOT_ENOUGH_FANS = 22
+            STATUS_NOT_ENOUGH_NEIGHBOURS = 23
+            STATUS_NO_PEAK_RADIO = 24
+            STATUS_RADIO_NOT_FOUND = 25
+            STATUS_API_KEY_SUSPENDED = 26
+            STATUS_DEPRECATED = 27
+            STATUS_RATE_LIMIT_EXCEEDED = 29
         """
 
         return self.status
@@ -2219,11 +2229,18 @@ class User(_BaseObject, _Chartable):
 
     def get_artist_tracks(self, artist, cacheable=False):
         """
+        Deprecated by Last.fm.
         Get a list of tracks by a given artist scrobbled by this user,
         including scrobble time.
         """
-        # Not implemented:
-        # "Can be limited to specific timeranges, defaults to all time."
+
+        warnings.warn(
+            "User.get_artist_tracks is deprecated and will be removed in a future "
+            "version. User.get_track_scrobbles is a partial replacement. "
+            "See https://github.com/pylast/pylast/issues/298",
+            DeprecationWarning,
+            stacklevel=2,
+        )
 
         params = self._get_params()
         params["artist"] = artist
@@ -2333,7 +2350,7 @@ class User(_BaseObject, _Chartable):
 
         params = self._get_params()
         if limit:
-            params["limit"] = limit
+            params["limit"] = limit + 1  # in case we remove the now playing track
         if time_from:
             params["from"] = time_from
         if time_to:
@@ -2341,7 +2358,11 @@ class User(_BaseObject, _Chartable):
 
         seq = []
         for track in _collect_nodes(
-            limit, self, self.ws_prefix + ".getRecentTracks", cacheable, params
+            limit + 1 if limit else None,
+            self,
+            self.ws_prefix + ".getRecentTracks",
+            cacheable,
+            params,
         ):
 
             if track.hasAttribute("nowplaying"):
@@ -2357,6 +2378,9 @@ class User(_BaseObject, _Chartable):
                 PlayedTrack(Track(artist, title, self.network), album, date, timestamp)
             )
 
+        if limit:
+            # Slice, in case we didn't remove a now playing track
+            seq = seq[:limit]
         return seq
 
     def get_country(self):
@@ -2514,6 +2538,32 @@ class User(_BaseObject, _Chartable):
 
         return self._get_things("getTopTracks", "track", Track, params, cacheable)
 
+    def get_track_scrobbles(self, artist, track, cacheable=False):
+        """
+        Get a list of this user's scrobbles of this artist's track,
+        including scrobble time.
+        """
+
+        params = self._get_params()
+        params["artist"] = artist
+        params["track"] = track
+
+        seq = []
+        for track in _collect_nodes(
+            None, self, self.ws_prefix + ".getTrackScrobbles", cacheable, params
+        ):
+            title = _extract(track, "name")
+            artist = _extract(track, "artist")
+            date = _extract(track, "date")
+            album = _extract(track, "album")
+            timestamp = track.getElementsByTagName("date")[0].getAttribute("uts")
+
+            seq.append(
+                PlayedTrack(Track(artist, title, self.network), album, date, timestamp)
+            )
+
+        return seq
+
     def get_image(self, size=SIZE_EXTRA_LARGE):
         """
         Returns the user's avatar
@@ -2736,7 +2786,19 @@ def _collect_nodes(limit, sender, method_name, cacheable, params=None):
 
     while not end_of_pages and (not limit or (limit and len(nodes) < limit)):
         params["page"] = str(page)
-        doc = sender._request(method_name, cacheable, params)
+
+        tries = 1
+        while True:
+            try:
+                doc = sender._request(method_name, cacheable, params)
+                break  # success
+            except Exception as e:
+                if tries >= 3:
+                    raise e
+                # Wait and try again
+                time.sleep(1)
+                tries += 1
+
         doc = cleanup_nodes(doc)
 
         # break if there are no child nodes
@@ -2810,8 +2872,9 @@ def _extract_top_albums(doc, network):
         name = _extract(node, "name")
         artist = _extract(node, "name", 1)
         playcount = _extract(node, "playcount")
+        info = {"image": _extract_all(node, "image")}
 
-        seq.append(TopItem(Album(artist, name, network), playcount))
+        seq.append(TopItem(Album(artist, name, network, info=info), playcount))
 
     return seq
 
@@ -2844,7 +2907,7 @@ def _extract_tracks(doc, network):
 def _url_safe(text):
     """Does all kinds of tricks on a text to make it safe to use in a URL."""
 
-    return url_quote_plus(url_quote_plus(_string(text))).lower()
+    return quote_plus(quote_plus(_string(text))).lower()
 
 
 def _number(string):
